﻿Imports System.Runtime.InteropServices

Class AudioRecPlay

    ' ==================================================================================================
    '  The AudioRecPlay class
    ' ==================================================================================================
    ' This class simplifies the following functions:
    ' - Record WAV files
    ' - Convert WAV files to MP3
    ' - Set WAV and MP3 recording properties
    ' - Set play Position in milliseconds
    ' - Get Position and Length  in milliseconds
    ' - Get Mode and audio Level
    ' - Set Volume and Balance (WAV and MP3 only) 
    ' - List and select Input Devices
    ' - Play WAV, MP3 and MIDI files

    ' With this class all the operations are simplified.
    ' All difficult tasks are solved in the class.
    ' Long path names are no more a problem.
    ' The parameters are passed to the MCI functions taking into account all their oddities.

    ' ==================================================================================================
    '  MCI capabilities and notes
    ' ==================================================================================================
    ' --------------------------------------------------------------------------------------------------
    ' MCI can play .WAV and .MID files,
    ' --------------------------------------------------------------------------------------------------
    ' MCI can not play files with .MP3 extension. 
    ' MCI just does not understand what is the file with .MP3 extension.
    ' MCI can play a .WAV file with MP3 compressed data inside.
    ' MCI opens .WAV file, reads its header, looks for the compression used and selects the right codec. 
    ' --------------------------------------------------------------------------------------------------

    ' --------------------------------------------------------------------------------------------------
    ' To make MCI playing MP3 files (with type WaveAudio)
    ' --------------------------------------------------------------------------------------------------
    ' mciSendString("open """ + filename + """ type WaveAudio alias " + name, "", 0, 0)
    ' --------------------------------------------------------------------------------------------------

    ' --------------------------------------------------------------------------------------------------
    ' To make MCI playing MP3 files (with REGISTRY)
    ' --------------------------------------------------------------------------------------------------
    ' Computer\HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows NT\CurrentVersion\MCI Extensions
    ' Change Key "mp3" from "MPEGVideo" to "WaveAudio"
    ' --------------------------------------------------------------------------------------------------

    ' --------------------------------------------------------------------------------------------------
    ' To make MCI playing MP3 files and saving all type files (path too long)
    ' --------------------------------------------------------------------------------------------------
    ' If the path is too long then MP3 files are not played (error MCIERR_FILENAME_REQUIRED)
    ' If the path is too long then all files are not saved (error MCIERR_FILENAME_REQUIRED)
    ' Change the extenson to WAV or use the GetShortPathName function

    ' --------------------------------------------------------------------------------------------------
    '  The MP3 conversion is NOT IMPLEMENTED
    ' --------------------------------------------------------------------------------------------------
    '  Recorded files are always PCM (also if the extension i MP3)
    '  To obtain true MP3 files a second pass with calls to LAME compressor is required


    ' ==================================================================================================
    '  PRIVATE FUNCTIONS
    ' ==================================================================================================
    Private Declare Function mciSendString Lib "winmm.dll" Alias "mciSendStringA" _
                            (ByVal lpstrCommand As String, _
                             ByVal lpstrReturnString As String, _
                             ByVal uReturnLength As Integer, _
                             ByVal hwndCallback As Integer) As Integer

    Private Declare Function GetShortPathName Lib "kernel32" Alias "GetShortPathNameA" _
                            (ByVal lpszLongPath As String, _
                             ByVal lpszShortPath As String, _
                             ByVal cchBuffer As Integer) As Integer

    ' This version works on existent files (path and filename are shortened)
    Private Shared Function GetShortPathName_Load(ByVal LongPathName As String) As String
        Dim ShortPathName As String = Space(256)
        Dim ret As Int32
        ret = GetShortPathName(LongPathName, ShortPathName, 256)
        Return Left(ShortPathName, ret)
    End Function

    ' working for non existent files (path shortened + filename unshorted)
    Private Shared Function GetShortPathName_Save(ByVal LongPathName As String) As String
        Dim FileName As String = IO.Path.GetFileName(LongPathName)
        Dim PathName As String = IO.Path.GetDirectoryName(LongPathName)
        Dim ShortPathName As String = Space(256)
        Dim ret As Int32
        ret = GetShortPathName(PathName, ShortPathName, 256)
        Return Left(ShortPathName, ret) + "\" + FileName
    End Function


    Private MediaName As String
    Private Shared InstanceID As Int32
    Private Shared LevelMeterStereo As Boolean

    ' ==================================================================================================
    '  PUBLIC SHARED FUNCTIONS
    ' ==================================================================================================
    Friend Sub New()
        InstanceID += 1
        MediaName = "media" + Process.GetCurrentProcess.Id.ToString + InstanceID.ToString
    End Sub

    Friend Shared Sub StartLevelMeter(Optional ByVal source As Int32 = -1, Optional ByVal stereo As Boolean = False)
        LevelMeterStereo = stereo
        mciSendString("close wavelevel", "", 0, 0)
        mciSendString("open new type waveaudio alias wavelevel", "", 0, 0)
        If stereo Then
            mciSendString("set wavelevel alignment 4 channels 2 bitspersample 16 samplespersec 22050 bytespersec 88200", "", 0, 0)
        Else
            mciSendString("set wavelevel alignment 2 channels 1 bitspersample 16 samplespersec 22050 bytespersec 44100", "", 0, 0)
        End If
        mciSendString("set wavelevel input " + source.ToString, "", 0, 0)
    End Sub

    Friend Shared Function GetLevel() As Single
        Dim Level As String = Space(256)
        mciSendString("status wavelevel level", Level, 256, 0)
        Return CSng(Val(Level) / 32768) ' Level returned is 0 to 1
    End Function

    ' To use stereo levels the "StartLevelMeter" function must be called with the "stereo" flag
    Friend Shared Function GetLevelLeft() As Single
        Dim Level As String = Space(256)
        mciSendString("status wavelevel level", Level, 256, 0)
        Return CSng((CInt(Val(Level)) And 65535) / 32768) ' Level returned is 0 to 1
    End Function
    Friend Shared Function GetLevelRight() As Single
        Dim Level As String = Space(256)
        mciSendString("status wavelevel level", Level, 256, 0)
        Return CSng((CInt(Val(Level)) >> 16) / 32768) ' Level returned is 0 to 1
    End Function

    ' This function gets the length of a closed file
    Friend Shared Function GetFileLengthMillisec(ByVal filename As String) As Int32
        filename = GetShortPathName_Load(filename)
        mciSendString("close getfilelength", "", 0, 0)
        Dim err As Int32
        err = mciSendString("open """ + filename + """ type waveaudio alias getfilelength", "", 0, 0)
        If err <> 0 Then
            err = mciSendString("open """ + filename + """ type mpegvideo alias getfilelength", "", 0, 0)
        End If
        Dim Length As String = Space(256)
        mciSendString("status getfilelength length", Length, 256, 0)
        mciSendString("close getfilelength", "", 0, 0)
        Return CInt(Val(Length))
    End Function


    ' ==================================================================================================
    '  PUBLIC FUNCTIONS
    ' ==================================================================================================
    ' -----------------------------------------------------------------
    '  All the format params must be sent with a single mciSendString
    '  SamplesPerSec = 100 to 98000
    ' -----------------------------------------------------------------
    Friend Sub SetFormat(ByVal Stereo As Boolean, _
                          ByVal Bits16 As Boolean, _
                          ByVal SamplesPerSec As Int32)
        Dim Format As String
        Dim Channels As Int32 = If(Stereo, 2, 1)
        Dim BitsPerSample As Int32 = If(Bits16, 16, 8)
        Dim Align As Int32 = If(Stereo, 2, 1) * If(Bits16, 2, 1)
        Format = " alignment " + Align.ToString
        Format += " channels " + Channels.ToString
        Format += " bitspersample " + BitsPerSample.ToString
        Format += " samplespersec " + SamplesPerSec.ToString
        Format += " bytespersec " + (Align * SamplesPerSec).ToString
        Dim n As Int32 = mciSendString("set " + MediaName + Format, "", 0, 0)
    End Sub

    Friend Sub Record(ByVal Stereo As Boolean, _
                      ByVal Bits16 As Boolean, _
                      ByVal SamplesPerSec As Int32, _
                      Optional ByVal source As Int32 = -1)

        mciSendString("close " + MediaName, "", 0, 0)
        mciSendString("open new type waveaudio alias " + MediaName, "", 0, 0)
        mciSendString("set " + MediaName + " time format milliseconds", "", 0, 0)

        SetFormat(Stereo, Bits16, SamplesPerSec)

        mciSendString("set " + MediaName + " input " + source.ToString, "", 0, 0)

        mciSendString("record " + MediaName, "", 0, 0)

        ' --------------------------------------------- wait up to 2 seconds for the recording startup
        For timeout As Int32 = 1 To 20
            Threading.Thread.Sleep(100)
            If IsRecording() Then Exit For
        Next
    End Sub
    Friend Sub [Stop]()
        mciSendString("stop " + MediaName, "", 0, 0)
    End Sub
    Friend Sub Save(ByVal filename As String)
        filename = GetShortPathName_Save(filename)
        mciSendString("save " + MediaName + " """ + filename + """", "", 0, 0)
        mciSendString("close " + MediaName, "", 0, 0)
    End Sub

    Friend Sub Close()
        mciSendString("close " + MediaName, "", 0, 0)
    End Sub

    Friend Sub OpenFile(ByVal filename As String)
        filename = GetShortPathName_Load(filename)
        mciSendString("close " + MediaName, "", 0, 0)
        Dim err As Int32
        err = mciSendString("open """ + filename + """ type waveaudio alias " + MediaName, "", 0, 0)
        If err = 0 Then Return
        err = mciSendString("open """ + filename + """ type mpegvideo alias " + MediaName, "", 0, 0)
    End Sub
    Friend Sub Play(Optional ByVal FromMillisec As Int32 = 0, _
                    Optional ByVal ToMillisec As Int32 = 0)
        Dim fromto As String = ""
        If FromMillisec > 0 Then fromto += " from " + FromMillisec.ToString
        If ToMillisec > 0 Then fromto += " to " + ToMillisec.ToString
        mciSendString("play " + MediaName + fromto, "", 0, 0)
    End Sub

    Friend Function GetPosition() As Int32
        Dim Position As String = Space(256)
        mciSendString("status " + MediaName + " position", Position, 256, 0)
        Return CInt(Val(Position))
    End Function
    Friend Function GetLength() As Int32
        Dim Length As String = Space(256)
        mciSendString("status " + MediaName + " length", Length, 256, 0)
        Return CInt(Val(Length))
    End Function

    ' Returns: "playing", "paused", "stopped" or "not ready"
    Friend Function GetMode() As String
        Dim Mode As String = Space(256)
        mciSendString("status " + MediaName + " mode", Mode, 256, 0)
        Return Mode.Replace(Chr(0), "").Trim
    End Function
    Friend Function IsPlaying() As Boolean
        Return GetMode() = "playing"
    End Function
    Friend Function IsPaused() As Boolean
        Return GetMode() = "paused"
    End Function
    Friend Function IsStopped() As Boolean
        Return GetMode() = "stopped"
    End Function
    Friend Function IsReady() As Boolean
        Return GetMode() <> "not ready"
    End Function
    Friend Function IsRecording() As Boolean
        Return GetMode() = "recording"
    End Function

    Friend Sub SetVolume(ByVal volume As Int32)
        mciSendString("setaudio " + MediaName + " volume to " + volume.ToString, "", 0, 0)
    End Sub
    Public Sub SetBalance(ByVal balance As Int32)
        mciSendString("setaudio " + MediaName + " left volume to " + (1000 - balance).ToString, "", 0, 0)
        mciSendString("setaudio " + MediaName + " right volume to " + balance.ToString, "", 0, 0)
    End Sub

    Friend Sub Pause()
        mciSendString("pause " + MediaName, "", 0, 0)
    End Sub
    Friend Sub [Resume]()
        mciSendString("resume " + MediaName, "", 0, 0)
    End Sub

    ' --------------------------------------------------------------------------- Seek repetition limited every 50 mS 
    Private WithEvents SeekTimer As Timer = New Timer
    Private SeekRepetitionLimitedMS As Int32
    Friend Sub SeekAndPlay_RepetitionLimited(ByVal millisec As Int32)
        Me.Stop()
        SeekRepetitionLimitedMS = millisec
        SeekTimer.Stop()
        SeekTimer.Start()
        SeekTimer.Interval = 50
    End Sub
    Private Sub SeekTimer_Tick(ByVal sender As Object, ByVal e As System.EventArgs) Handles SeekTimer.Tick
        SeekTimer.Stop()
        Seek(SeekRepetitionLimitedMS)
        Play()
    End Sub
    ' --------------------------------------------------------------------------- Seek normal (immediately executed) 
    Friend Sub Seek(ByVal millisec As Int32)
        Dim TotalMS As Int32 = GetLength()
        If millisec > TotalMS Then millisec = TotalMS
        mciSendString("seek " + MediaName + " to " + millisec.ToString, "", 0, 0)
    End Sub
    Friend Sub SeekToStart()
        mciSendString("seek " + MediaName + " to start", "", 0, 0)
    End Sub
    Friend Sub SeekToEnd()
        mciSendString("seek " + MediaName + " to end", "", 0, 0)
    End Sub
    '  Working only if opened with waveaudio not with mpegvideo or midi
    'Friend Sub SetDestination(ByVal dest As Int32)
    '    Dim n As Int32 = mciSendString("set " + MediaName + " output " + dest.ToString, "", 0, 0)
    'End Sub


    '=======================================================================
    ' WAV to MP3
    '=======================================================================

    ' ----------------------------------------------------------------------
    ' COMPRESSION RATES
    ' ----------------------------------------------------------------------
    ' 60 min - 44100 - WAV = 635 mega / MP3 128kbps = 58 mega (BP 22k)
    ' 60 min - 44100 - WAV = 635 mega / MP3  96kbps = 43 mega (BP 16k)
    ' 60 min - 22050 - WAV = 317 mega / MP3  64kbps = 28 mega (BP 10k)
    ' 60 min - 11025 - WAV = 160 mega / MP3  32kbps = 14 mega (BP  6k)
    ' 60 min -  5500 - WAV =  80 mega / MP3  16kbps =  7 mega (BP  3k)

    ' ----------------------------------------------------------------------
    ' BITRATE
    ' ----------------------------------------------------------------------
    ' -b n  = set constant bitrate
    '        (8 16 24 32 40 48 56 64 80 96 112 128 144 160 192 224 256 320)

    ' ----------------------------------------------------------------------
    ' MODE
    ' ----------------------------------------------------------------------
    '-m m    mono
    '-m l    get only the left channel of a stereo signal for a mono output
    '-m r    get only the right channel of a stereo signal for a mono output
    '-m s    (forced) L/R stereo
    '-m j    joint stereo
    '-m f    forced mid/side stereo
    '-m d    dual (independent) channels (used for dual languages)

    ' ----------------------------------------------------------------------
    ' QUALITY
    ' ----------------------------------------------------------------------
    '-q 0 =	Use slowest & best possible version of all algorithms.
    '-q 3 = Default value. Good speed, good quality
    '-q 7 = Very fast, ok quality. (psycho acoustics are used for pre-echo & M/S, but no noise shaping is done.
    '-q 9 = Disables almost all algorithms including psy-model. Poor quality.

    ' ----------------------------------------------------------------------
    ' LOW PASS FREQUENCY
    ' ----------------------------------------------------------------------
    '--lowpass n = lowpass filter freq. 50 to 50000 Hz
    ' (normally unused) 
    ' (use only to increase a little the bandpass for 16, 32 and 64 kbps) 

    ' ----------------------------------------------------------------------
    ' PRIORITY - Windows and OS/2 process priority control
    ' ----------------------------------------------------------------------
    '--priority 0 = Low priority (IDLE, delta = 0)
    '--priority 1 = Medium priority (IDLE, delta = +31)
    '--priority 2 = Regular priority (REGULAR, delta = -31)
    '--priority 3 = High priority (REGULAR, delta = 0)
    '--priority 4 = Maximum priority (REGULAR, delta = +31)

    ' ----------------------------------------------------------------------
    ' Conversion time
    ' ----------------------------------------------------------------------
    '  WAV time     --  Conversion time --
    '               Quality 3   Quality 7
    ' ------------------------------------------
    '  1 minute  =   3.3 sec     2.7 sec
    '  3 minutes =   7 sec       5 sec
    ' 10 minutes =  32 sec      15 sec
    ' 30 minutes =  90 sec ?    40 sec ?

    Friend Sub WavToMp3(ByVal WavFile As String, _
                        Optional ByVal Bitrate As Int32 = 128, _
                        Optional ByVal Stereo As Boolean = True, _
                        Optional ByVal Quality As Int32 = 3, _
                        Optional ByVal LowPassFreq As Int32 = 0, _
                        Optional ByVal Priority As Int32 = 2, _
                        Optional ByVal ProgressCallbackFunction As Action(Of Single) = Nothing)

        ' ------------------------------------------------------------------ start lame converter
        Dim Mp3File As String = IO.Path.ChangeExtension(WavFile, ".mp3")
        IO.File.Delete(Mp3File)

        Dim proc As New Process()
        proc.StartInfo.Arguments = "-b " + Bitrate.ToString + " "
        If Not Stereo Then proc.StartInfo.Arguments += " -m m "
        proc.StartInfo.Arguments += "-q " + Quality.ToString + " "
        proc.StartInfo.Arguments += "--priority " + Priority.ToString + " "
        If LowPassFreq >= 50 And LowPassFreq <= 50000 Then
            proc.StartInfo.Arguments += "--lowpass " + LowPassFreq.ToString("0") + " "
        End If
        proc.StartInfo.Arguments += """" + WavFile + """ """ + Mp3File + """"
        proc.StartInfo.FileName = Application.StartupPath + "\lame.exe"
        proc.StartInfo.WindowStyle = ProcessWindowStyle.Hidden
        proc.Start()

        ' ------------------------------------------------------------------ wait and callback for progress
        Dim WavFileMilliseconds As Int32 = GetFileLengthMillisec(WavFile)
        Dim PredictedFileLength As Single = WavFileMilliseconds * Bitrate / 8.0F + 1024
        Dim Mp3FileLength As Long
        Do
            If ProgressCallbackFunction IsNot Nothing AndAlso PredictedFileLength > 0 Then
                If IO.File.Exists(Mp3File) Then
                    Mp3FileLength = My.Computer.FileSystem.GetFileInfo(Mp3File).Length
                End If
                ProgressCallbackFunction(Mp3FileLength / PredictedFileLength)
            End If
            Threading.Thread.Sleep(100)
        Loop Until proc.HasExited

    End Sub


    '=======================================================================
    ' WAVE IN DEVICES
    '=======================================================================
    <StructLayout(LayoutKind.Sequential)> _
    Private Structure WAVEINCAPS
        Dim wMid As Int16
        Dim wPid As Int16
        Dim vDriverVersion As Int32
        <MarshalAs(UnmanagedType.ByValTStr, SizeConst:=32)> Public szPname As String
        Dim dwFormats As Int32
        Dim wChannels As Int16
        Dim wReserved1 As Int16
    End Structure

    Private Declare Function waveInGetNumDevs Lib "winmm.dll" () As Int32

    ' DeviceID is a value from 1 to wInGetNumDevs (0 = default device)
    Private Declare Function waveInGetDevCaps Lib "winmm.dll" Alias "waveInGetDevCapsA" _
                                                            (ByVal DeviceID As Int32, _
                                                             ByRef Caps As WAVEINCAPS, _
                                                             ByVal Size As Int32) As Int32

    Friend Shared Function GetDevNames() As String()
        Dim Names(waveInGetNumDevs() - 1) As String
        Dim wcap As WAVEINCAPS = New WAVEINCAPS
        For i As Int32 = 0 To waveInGetNumDevs() - 1
            waveInGetDevCaps(i, wcap, Marshal.SizeOf(wcap))
            Names(i) = wcap.szPname.Trim
        Next
        Return Names
    End Function

End Class



